/*
** Module   : AdminClient.java
** Abstract : Applet used as a front-end for the P2J administration client.
**
** Copyright (c) 2009-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 JJC 20090107  @41134  Created the first version of the file.
** 002 NVS 20090304  @41403  Reworked and brought to compliance with the
**                           P2J coding standards. Renamed.
** 003 NVS 20090312  @41524  Implemented menu system, swappable screens
**                           and the logon screen.
** 004 NVS 20090312  @41622  Refactored the client's code into multiple
**                           classes.
** 005 NVS 20090318  @41630  Javadoc changes
** 006 NVS 20090323  @41664  Refactored one more time to get rid of static
**                           members.
** 007 NVS 20090326  @41688  Added Help --> Show errors menu item and the
**                           relevant backing code.
** 008 NVS 20090428  @42036  Added target change notification framework.
** 009 NVS 20090429  @42078  Added bracketing of server calls with busy/idle
**                           mouse cursors.
** 010 CA  20090513  @42176  Added getter for the AdminMenu instance.
** 011 CA  20090522  @42431  Exposed alarm/ok methods to be available from
**                           customer-specific plugins.
** 012 CA  20090529  @42488  Added getter for the net session. Exposed beep 
**                           method to be available from plugins.
** 013 CA  20090604  @42567  Added getAccountExtPlugin.
** 014 CA  20090720  @43242  Added getCustomAclEditor.
** 015 NVS 20090723  @43358  Added useful showMessages() variations.
** 016 NVS 20090724  @43390  Added getResourceTypes() method (delegate).
** 017 NVS 20090729  @43427  Implemented centralized printing.
** 018 NVS 20090729  @434    A better print error reporting through a dialog.
** 019 NVS 20090803  @43480  Using a better approach to printing with extended
**                           detail choice. Reusable printing methods are
**                           implemented here.
** 020 NVS 20090806  @43512  Fixed NPE caused by missing map allocation.
** 021 NVS 20090807  @43548  Refactored in support of custom menu extensions.
** 022 NVS 20091007  @44111  Implemented LOGON notification broadcasting.
** 023 SIY 20091116  @44341  Minor cleanups and refactoring.
** 024 SIY 20091211  @44507  Changed visibility of add(JPanel, String, 
**                           PrintRequest).
** 025 SIY 20100419  @44825  Implemented automatic adding of function into
**                           directory if it does not exists.
** 026 GES 20111003          Slight change to explicitly specify the
**                           authentication processing.
** 027 OM  20121002          Added NPE guard for menuExt in postLogon()
*/

package com.goldencode.p2j.admin.client;

import com.goldencode.p2j.cfg.BootstrapConfig;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.security.RightsEditor;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.admin.AdminExports;
import com.goldencode.p2j.admin.AdminDef;

import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Color;
import java.awt.BorderLayout;
import java.awt.Frame;
import java.awt.Toolkit;
import java.awt.event.*;
import java.awt.print.PrinterException;
import java.text.MessageFormat;
import java.util.*;
import javax.swing.*;
import javax.swing.table.*;

/**
 * This class initializes the admin client session and provides screen
 * switching services for the rest of the menu driven code.
 *
 * @author NVS
 */
public class AdminClient 
extends JApplet
implements AdminNotify, PrintRequest
{
   /** Screen switcher - empty screen ID */
   public static final String SCREEN_BLANK = "adm_blank";

   /** Screen switcher - logon screen ID */
   public static final String SCREEN_LOGON = "adm_logon";
   
   /** Screen switcher - errors screen ID */
   public static final String SCREEN_ERRORS = "adm_errors";
   
   /** info line item */
   public static final int TARGET_SERVER = 0;
   
   /** info line item */
   public static final int TARGET_TARGET = 1;
   
   /** info line item */
   public static final int TARGET_ERROR = 2;
   
   /** instance of Frame suitable for creating dialog */
   private Frame frame = null;
   
   /** Related instance of AdminMenu */
   private AdminMenu menu = null;

   /** Related instance of AdminLogon */
   private AdminLogon logon = null;

   /** Related instance of AdminTarget */
   private AdminTarget target = null;

   /** Admin server host */
   private String host = null;

   /** Admin server port */
   private String port = null;

   /** Current admin server display name (variable) */
   private String targetServer = null;
   
   /** P2J server name */
   @SuppressWarnings("unused")
   private String serverName = null;
   
   /** Admin Server APIs */
   private AdminExports adm = null;

   /** error messages list data model */
   private DefaultListModel errData = null;

   /** list to display error messages */
   private JList errBox = null;
   
   /** scroller for error messages */
   private JScrollPane errScrollPane = null;
   
   /** map of printable screens */
   private Map<String, PrintRequest>printers = 
                                          new HashMap<String, PrintRequest>();
   /** 
    * BootstrapConfig used in applet.
    * A single instance should be used for multiple consecutive sessions,
    * because the SecurityManager caches the instance during creation.  
    */
   private BootstrapConfig bc = null;
      
   /**
    * Instance of SecurityManager used in applet.
    * A single instance should be used for multiple consecutive sessions.
    */
   @SuppressWarnings("unused")
   private SecurityManager sm = null;

   /**
    * This is the main applet panel, made of the title, swapper and the target
    * components.
    */
   private JPanel mainPanel = null;

   /** Swapper panel */
   private JPanel swapPanel = null;
   
   /** Swapper layout manager */
   private CardLayout screens = null;
   
   /** Current screen title */
   private JLabel title = null;
   
   /** Name of the currently displayed screen */
   private String showing = null;
   
   /** Admin session target identification */ 
   @SuppressWarnings("unused")
   private JLabel targetInfo = null;
   
   /** target server P2J server name */ 
   private JLabel tgtServer = null;
   
   /** target server P2J server name separator */ 
   private JLabel tgtSrvSep = null;
   
   /** target server host name */ 
   private JLabel tgtHost = null;
   
   /** target server port name */ 
   private JLabel tgtPort = null;
   
   /** target server's target name live/file etc */ 
   private JLabel tgtTarget = null;
   
   /** target server's last error message */ 
   private JLabel tgtError = null;
   
   /** logon listeners registry */
   private Set<AdminNotify> logonList = null;

   /** target change listeners registry */
   private Set<AdminNotify> targetList = null;
   
   /** logoff listeners registry */
   private Set<AdminNotify> logoffList = null;
   
   /** flag telling messages are showing */
   private boolean inMessages = false;
   
   /** flag telling print request being processed */
   private boolean inPrint = false;
   
   /** current print request processor */
   private PrintRequest currentPrinter = null;

   /** map of last used filters per by screen ID */
   private Map<String, String>lastFilters = null;

   /** the current admin definition values. */
   private AdminDef adminDef = null;
   
   /** the instance of the account extension plugin */
   private ClientAccountExtension accountExt = null;
   
   /** the instance of the menu extension plugin */
   private AdminCustomMenu menuExt = null;
   
   /** the custom ACL editors. */
   private Map<String, RightsEditor> customAclEditors = null;
   
   /** resource types */
   private String[] resourceTypes = null;

   static
   {
      try
      {
         UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }
   
//   public AdminClient() 
//   {
//      super();
//   }
   
   /**
    * Get the current P2J session.
    * 
    * @return   current P2J session.
    */
   public Session getSession()
   {
      return logon.getSession();
   }
   
   /** 
    * Prints the current screen by delegating the request to a proper registered
    * print request processor.
    */
   void printScreen()
   {
      if (inMessages)
      {
         // server messages screen is printed right from here
         inPrint = true;
         currentPrinter = this;
         print();
         inPrint = false;
         return;
      }
      
      // get the printer for the current screen
      PrintRequest pr = printers.get(showing);
      
      // dispatch the request
      if (pr == null)
      {
         AdminUtils.errorDialog(this, "Print Error", "nothing to print here");
      }
      else
      {
         inPrint = true;
         currentPrinter = pr;
         pr.print();
         inPrint = false;
      }
   }

   /**
    * Performs printing of the current screen data.
    * This screen prints the current view of the server messages.
    */
   public void print()
   {
      // build a temporary table
      int nall = errData.size();
      
      if (nall == 0)
      {
         AdminUtils.errorDialog(this, "Print Error", "nothing to print here");
         return;
      }
      
      String[] all = new String[nall];
      
      for (int i = 0; i < nall; i++)
      {
         // get all messages
         all[i] = (String)errData.get(i);
      }
      
      TableModel tm = new PrintMessagesModel(all);
      JTable table = new JTable(tm);
      table.setShowHorizontalLines(false);
      
      MessageFormat header = new MessageFormat("Server Messages");
      MessageFormat footer = new MessageFormat("page {0, number}");
      
      String savedShowing = showing;
      showing = SCREEN_ERRORS;
      
      printOn(table, tm, header, footer);
      
      showing = savedShowing;
   }
   
   /**
    * Queries the extended printing capability. If the extended capability is
    * not provided, the printExtended() method should not be called.
    * 
    * @return   <code>false</code> to tell the caller the extended printing is
    *           not implemented.
    */
   public boolean canPrintExtended()
   {
      return false;
   }

   /**
    * Does nothing here.
    */
   public void printExtended()
   {
   }
   
   /**
    * Puts the specified panel on top temporarily for printing.
    * 
    * @param    table
    *           a table that serves as a print preview
    * 
    * @param    tm
    *           table model for the given table
    * 
    * @param    header
    *           this header will be used for table printout
    * 
    * @param    footer
    *           this footer will be used for table printout
    */
   void printOn(final JTable table, TableModel tm, final MessageFormat header, 
                final MessageFormat footer)
   {
      // safety
      if (!inPrint)
      {
         return;
      }
      
      // attach a sorter to the table
      final TableRowSorter<TableModel> sorter = new TableRowSorter<TableModel>(tm);
      sorter.setRowFilter(null);
      table.setRowSorter(sorter);
      
      // create a panel for the table inside of a scroller
      JScrollPane scroller = new JScrollPane();
      scroller.getViewport().add(table);
      JPanel tpanel = AdminUtils.borderComponents(new Component[] {scroller}, 
                                                  true);
      
      // create a panel for print control buttons
      final JCheckBox cExtend = new JCheckBox("show all details"); 
      JButton bPrint  = new JButton("Print");
      JButton bFilter = new JButton("Filter");
      JButton bCancel = new JButton("Cancel");
      
      JPanel bpanel = AdminUtils.groupComponents(
                      new Component[][]{{cExtend, bPrint, bFilter, bCancel}});
      
      if (!currentPrinter.canPrintExtended())
      {
         cExtend.setSelected(false);
         cExtend.setEnabled(false);
      }
      
      final JPanel panel = AdminUtils.groupComponents(new Component[][]
                                                      {{tpanel}, {bpanel}});

      // Print button event handler
      bPrint.addActionListener(new ActionListener() 
         {
            @SuppressWarnings("unused")
            public void actionPerformed(ActionEvent evt) 
            {
               // check to see whether all details are requested
               if (cExtend.isSelected())
               {
                  currentPrinter.printExtended();
                  return;
               }
               
               // print the temporary table
               try
               {
                  table.print(JTable.PrintMode.FIT_WIDTH, header, footer);
               }
               catch (PrinterException pe)
               {
                  beep("printer exception" + pe);
               }
               catch (SecurityException se)
               {
                  // nothing to do
               }
               finally
               {
                  printOff(panel);
               }
            }
         });
      
      // Filter button event handler
      bFilter.addActionListener(new ActionListener() 
         {
            @SuppressWarnings({ "unused", "unchecked" })
            public void actionPerformed(ActionEvent evt) 
            {
               // get the last used filter for this screen, if any
               String lastFilter = lastFilters.get(showing);
               
               int    type = JOptionPane.QUESTION_MESSAGE;
               String txt  = 
                    "Define the filtering specification below. Press the\n" +
                    "'OK' button without any specification to remove any\n" +
                    "previous filter.";
               String spec = (String) JOptionPane.showInputDialog(
                                                  AdminClient.this,
                                                  txt,
                                                  "Filter",
                                                  type,
                                                  null,
                                                  null,
                                                  lastFilter);
               
               if (spec == null || spec.trim().length() == 0)
               {
                  // remove filtering
                  sorter.setRowFilter(null);
                  lastFilters.remove(showing);
               }
               else
               {
                  // install a filter
                  sorter.setRowFilter(AdminUtils.createFilter(spec));
                  lastFilters.put(showing, spec);
               }
            }
         });
      
      // Cancel button event handler
      bCancel.addActionListener(new ActionListener() 
         {
            @SuppressWarnings("unused")
            public void actionPerformed(ActionEvent evt) 
            {
               printOff(panel);
            }
         });
                                  
      // show the print preview panel
      menu.saveAndDisableMenu();
      
      swapPanel.add(panel, "print");
      screens.show(swapPanel, "print");
   }
   
   /**
    * Creates a temporary panel with the specified table and Print and Back
    * buttons that provide print and cancel functionality.
    * <p>
    * This method is supposed to be called after {@link #printOn} and before
    * {@link #printOff}, which means a print preview panel is showing by the
    * time of this call. When Print button is pressed, this method removes
    * both the extended and the regular print preview panels after printing is
    * done.
    * 
    * @param    table
    *           a table to print as an extended view
    * 
    * @param    header
    *           this header will be used for table printout
    * 
    * @param    footer
    *           this footer will be used for table printout
    */
   void printExtendedOn(final JTable table, final MessageFormat header, 
                        final MessageFormat footer) 
   {
      // create a panel for the table inside of a scroller
      JScrollPane scroller = new JScrollPane();
      scroller.getViewport().add(table);
      JPanel tpanel = AdminUtils.borderComponents(new Component[] {scroller}, 
                                                  true);
 
      // create a panel for two print control buttons
      JButton bPrint = new JButton("Print");
      JButton bBack  = new JButton("Back");
      
      JPanel bpanel = AdminUtils.groupComponents(
                      new Component[][]{{bPrint, bBack}});
      
      // create a final panel for both
      final JPanel panel = AdminUtils.groupComponents(new Component[][]
                                                      
                                                      {{tpanel}, {bpanel}});

      // Print button event handler
      bPrint.addActionListener(new ActionListener() 
         {
            @SuppressWarnings("unused")
            public void actionPerformed(ActionEvent evt) 
            {
               // print the extended view table
               try
               {
                  table.print(JTable.PrintMode.FIT_WIDTH, header, footer);
               }
               catch (PrinterException pe)
               {
                  beep("printer exception" + pe);
               }
               catch (SecurityException se)
               {
                  // nothing to do
               }
               finally
               {
                  // done with printing - reverse all temporary panels
                  printExtendedOff(panel);
                  printOff(panel);
               }
            }
         });
      
      // Back button event handler
      bBack.addActionListener(new ActionListener() 
         {
            @SuppressWarnings("unused")
            public void actionPerformed(ActionEvent evt) 
            {
               // remove the extended view panel and return to the regular one
               printExtendedOff(panel);
            }
         });
                                  
      // show the temporary extended view panel
      swapPanel.add(panel, "printExtended");
      screens.show(swapPanel, "printExtended");
   }
   
   /**
    * Reverses the action of the printExtendedOn() method.
    * 
    * @param    panel
    *           panel containing a table being printed
    */
   void printExtendedOff(JPanel panel)
   {
      screens.show(swapPanel, "print");
      screens.removeLayoutComponent(panel);
   }
   
   /**
    * Reverses the action of the printOn() method.
    * 
    * @param    panel
    *           panel containing a table being printed
    */
   void printOff(JPanel panel)
   {
      screens.show(swapPanel, showing);
      screens.removeLayoutComponent(panel);

      menu.restoreMenu();
   }
   
   /**
    * Adds an applet screen.
    * 
    * @param    panel
    *           main panel of the screen
    * 
    * @param    screenId
    *           identifier of the screen
    */
   public void add(JPanel panel, String screenId)
   {
      swapPanel.add(panel, screenId);
   }

   /**
    * Adds a printable applet screen.
    * 
    * @param    panel
    *           main panel of the screen
    * 
    * @param    screenId
    *           identifier of the screen
    * 
    * @param    printer
    *           <code>registers this screen as printable by the specified
    *           delegate.
    */
   public void add(JPanel panel, String screenId, PrintRequest printer)
   {
      // register the screen with the swapper
      add(panel, screenId);
      
      // register the printer
      printers.put(screenId, printer);
   }

   /**
    * Swaps the applet screen.
    * 
    * @param    screenId
    *           identifier of the desired screen
    */
   void show(String screenId)
   {
      screens.show(swapPanel, screenId);
      showing = screenId;
      inMessages = false;
   }

   /**
    * Swaps the applet screen and changes the title
    * 
    * @param    screenId
    *           identifier of the desired screen
    * 
    * @param    title
    *           new title
    */
   public void show(String screenId, String title)
   {
      screens.show(swapPanel, screenId);
      showing = screenId;
      this.title.setText(title);
      inMessages = false;
   }

   /**
    * Returns the reference to the account extension plugin.
    * 
    * @return   instance of the account extension plugin to use.
    */
   ClientAccountExtension getAccountExtPlugin()
   {
      return accountExt;
   }

   /**
    * Get the custom ACL editor for the given resource type;
    * 
    * @param    resourceType
    *           The resource type.
    * 
    * @return   The ACL editor for the given resource type.
    */
   RightsEditor getCustomAclEditor(String resourceType)
   {
      return customAclEditors.get(resourceType);
   }

   /**
    * Get the array of all resource types.
    * 
    * @return   resource types
    */
   String[] getResourceTypes()
   {
      return resourceTypes;
   }

   /**
    * Gets the target server identification.
    * 
    * @return   a string of description of the target sever
    */
   String getTargetServer()
   {
      return new String(targetServer);
   }

   /**
    * Sets the target server identification and displays it on the screen.
    * 
    * @param    item
    *           a constant identifying item to set; one of the TARGET_*
    *           definitions.
    * 
    * @param    display
    *           new text to display
    */
   void setTargetInfo(int item, String display)
   {
      switch(item)
      {
         case TARGET_SERVER:
            tgtServer.setText(display);
            tgtSrvSep.setText("@");
            break;
            
         case TARGET_TARGET:
            tgtTarget.setText(display);
            break;
            
         case TARGET_ERROR:
            tgtError.setText(display);
            break;
            
         default:
      }
   }
   
   /**
    * Receives the reference to the server admin exports.
    * 
    * @param    exports
    *           instance of the remote admin exports to use
    */
   void initExports(AdminExports exports)
   {
      adm = exports;
      
      // this signal also indicates successful logon
      tgtError.setText("");
   }

   /**
    * Returns the reference to the server admin exports.
    * 
    * @return   instance of the remote admin exports to use
    */
   public AdminExports getExports()
   {
      return adm;
   }
   
   /**
    * Receives the references to all items initialized by logon
    * 
    * @param    adminDef
    *           instance of AdminDef to cache
    * 
    * @param    resourceTypes
    *           array of resource type names
    * 
    * @param    accountExt
    *           custom account extension interface
    * 
    * @param    menuExt
    *           custom menu extension interface
    * 
    * @param    customAclEditors
    *           map of custom ACL editors per resource type
    */
   void postLogon(AdminDef adminDef, String[] resourceTypes, 
                  ClientAccountExtension accountExt, AdminCustomMenu menuExt, 
                  Map<String, RightsEditor> customAclEditors)
   {
      // save the references
      this.adminDef = adminDef;
      this.resourceTypes = resourceTypes;
      this.accountExt = accountExt;
      this.menuExt = menuExt;
      this.customAclEditors = customAclEditors;
      
      if (menuExt != null)
      {
         // rebuild the menu with the custom modifications
         List<JMenuItem> menuList = menu.getMenuList();
         menuExt.insertCustomMenuItems(menuList);
         menu.buildMenu(adminDef.profile);
         setJMenuBar(menu.getMenu());
      }      
      
      
      // notify all logon listeners
      for(AdminNotify al: logonList)
      {
         al.notifyEvent(AdminNotify.NOTIFY_LOGON);
      }
   }
   
   /**
    * Returns the instance of parent Frame.
    * 
    * @return   instance of parent Frame
    */
   Frame getFrame()
   {
      return frame;
   }
   
   /**
    * Returns the instance of AdminMenu.
    * 
    * @return   instance of AdminMenu
    */
   AdminMenu getMenu()
   {
      return menu;
   }
   
   /** 
    *  Initializes the applet within the browser.
    */
   public void init() 
   {
      // initialize the applet
      try 
      {
         // do it in a non-blocking way
         java.awt.EventQueue.invokeAndWait(new Runnable() 
         {
            public void run() 
            {
               initComponents();
            }            
         });
      } 
      catch (Exception ex) 
      {
         ex.printStackTrace();
         showStatus("init failure; see console");
      }
   }

   /** 
    *  Creates the UI components: menu, panels etc. and displays the logon
    *  panel as the starting point.
    */
   private void initComponents() 
   {
      // one time initialization
      logonList   = new HashSet<AdminNotify>();
      targetList  = new HashSet<AdminNotify>();
      logoffList  = new HashSet<AdminNotify>();
      lastFilters = new HashMap<String, String>();
      
      host = getParameter("p2jservername");
      port = getParameter("p2jserverport");
      
      if (host == null || host.length() == 0 ||
          port == null || port.length() == 0)
      {
         showStatus("target server configuration failure");
         return;
      }
      
      targetServer = new String(host + ":" + port);
      
      try
      { 
         bc = new BootstrapConfig();
         bc.setServer(false);
         bc.setConfigItem("security", "trust_mgr", "disable", "true");
         sm = SecurityManager.createInstance(bc);
      }
      catch (Exception ex)
      {
         ex.printStackTrace();
         showStatus("configuration failure; see console");
         return;
      }
      
      // locate a parent frame
      frame = findParentFrame();
      
      // instantiate additional classes
      logon  = new AdminLogon(bc, host, port);
      target = new AdminTarget();
      menu   = new AdminMenu();

      // cross initialize additional classes
      logon.init(this, menu, target);
      menu.init(this, logon, target);
      target.init(this, menu, logon);
      
      // create the menu system
      setJMenuBar(menu.getMenu());
      
      // disable everything but logon
      menu.disconnectedState();
      
      // create main panel for title, swapper and  info
      mainPanel = new JPanel(new BorderLayout());
      add(mainPanel);
      
      title = new JLabel();
      title.setHorizontalAlignment(SwingConstants.CENTER);
      title.setBackground(Color.DARK_GRAY);
      mainPanel.add(title, BorderLayout.NORTH);
      
      // information line structures
      Color green = (Color.GREEN).darker();
      JLabel tgtLabel = new JLabel("Target ");
      tgtServer = new JLabel();
      tgtServer.setForeground(green);
      tgtSrvSep = new JLabel("");
      tgtHost = new JLabel(host);
      JLabel tgtHstSep = new JLabel(":");
      tgtPort = new JLabel(port);
      JLabel tgtPrtSep = new JLabel("  ");
      tgtTarget = new JLabel(); 
      tgtTarget.setForeground(green);
      tgtError = new JLabel();
      tgtError.setForeground(Color.RED);
      tgtError.setHorizontalAlignment(SwingConstants.RIGHT);
      
      JPanel infoLine = new JPanel();
      infoLine.setLayout(new BoxLayout(infoLine, BoxLayout.LINE_AXIS));
      infoLine.add(tgtLabel);
      infoLine.add(tgtServer);
      infoLine.add(tgtSrvSep);
      infoLine.add(tgtHost);
      infoLine.add(tgtHstSep);
      infoLine.add(tgtPort);
      infoLine.add(tgtPrtSep);
      infoLine.add(tgtTarget);
      infoLine.add(Box.createHorizontalGlue());
      infoLine.add(tgtError);
       
      mainPanel.add(infoLine, BorderLayout.SOUTH);
      
      // create swapper panel
      screens = new CardLayout();
      swapPanel = new JPanel(screens);
      mainPanel.add(swapPanel, BorderLayout.CENTER);
      
      // swapper's blank screen
      JPanel blankPanel = new JPanel();
      swapPanel.add(blankPanel, SCREEN_BLANK);
      
      // swapper's logon screen
      JPanel lonPanel = logon.getLogonPanel();
      swapPanel.add(lonPanel, SCREEN_LOGON);
      
      // swapper's errors screen
      errorsScreen();
      
      // initiate logon
      logon.logonPanel();
   }

   /** 
    * Shows a short status message.
    * This method is of a very limited usefulness due to the fact the status
    * area of the hosting browser is shared among all multiple tabs.
    */
   void status(String message)
   {
      showStatus(message);
   }

   /** 
    * Displays the most recent error message and produces an alarm sound.
    */
   public void alarm()
   {
      // not connected yet
      if (adm == null)
      {
         beep("not connected");
         return;
      }
      
      // busy
      AdminUtils.busy(this);
      
      // get the message
      String error = adm.peekMessage();
      if (error == null)
      {
         error = "*** unknown error ***";
      }
      
      beep(error);
      
      // idle
      AdminUtils.idle(this);
   }

   /**
    * Places an error message with audible alarm.
    * 
    * @param    error
    *           error message text
    */
   public void beep(String error)
   {
      tgtError.setText(error);
      Toolkit.getDefaultToolkit().beep();
   }
   
   /** 
    * Clears the error message in info line.
    */
   public void ok()
   {
      tgtError.setText("");
   }
   
   /** 
    * Shows current server messages.
    */
   void showMessages()
   {
      // not connected yet
      if (adm == null)
      {
         beep("not connected");
         return;
      }
      
      // busy
      AdminUtils.busy(this);
      
      tgtError.setText("");
      
      // show the messages screen
      inMessages = true;
      screens.show(swapPanel, SCREEN_ERRORS);
      this.title.setText("Server messages");
      
      // get error messages and stuff the screen
      String[][] errs = adm.getMessages(false);

      for (int i = 0; i < errs.length; i++)
      {
         for (int j = 0; j < errs[i].length; j++)
         {
            errData.addElement(errs[i][j]);
         }
      }
      
      // conditionally enable clear messages function
      menu.setClearMessages(!errData.isEmpty());
      
      // idle
      AdminUtils.idle(this);
   }
   
   /**
    * Shows specified messages.
    * 
    * @param    msg
    *           array of messages
    */
   public void showMessages(String[] msg)
   {
      for (int i = 0; i < msg.length; i++)
      {
         errData.addElement(msg[i]);
      }
      
      // conditionally enable clear messages function
      menu.setClearMessages(!errData.isEmpty());
   }
   
   /**
    * Shows single message.
    * 
    * @param    msg
    *           message
    */
   public void showMessage(String msg)
   {
      showMessages(new String[] {msg});
   }
   
   /** 
    * Clears server messages.
    */
   void clearMessages()
   {
      // unconditionally disable itself
      menu.setClearMessages(false);
      
      // not connected yet
      if (adm == null)
      {
         beep("not connected");
         return;
      }

      // busy
      AdminUtils.busy(this);
      
      tgtError.setText("");
      
      // clear messages on the server
      adm.clearMessages();

      // clear message screen
      errData.removeAllElements();
      
      inMessages = true;
      screens.show(swapPanel, SCREEN_ERRORS);
      this.title.setText("Server messages");
      
      // idle
      AdminUtils.idle(this);
   }
   
   /**
    * Registers listeners for logon events.
    * 
    * @param    al
    *           instance of AdminNotify interface that should be notified with
    *           logon events
    */
   void registerLogonListener(AdminNotify al)
   {
      logonList.add(al);
   }
   
   /**
    * Registers listeners for logoff events.
    * 
    * @param    al
    *           instance of AdminNotify interface that should be notified with
    *           logoff events
    */
   void registerLogoffListener(AdminNotify al)
   {
      logoffList.add(al);
   }
   
   /**
    * Registers listeners for target change events.
    * 
    * @param    al
    *           instance of AdminNotify interface that should be notified with
    *           target change events
    */
   void registerTargetListener(AdminNotify al)
   {
      targetList.add(al);
   }
   
   /**
    * Notifies about successful logoff or target change and gives chance to
    * release resources.
    * <p>
    * This method receives the logoff notification from AdminLogon class and
    * then further notifies all registered listeners.
    * <p>
    * Target change notification comes from AdminTarget class. The reaction is
    * similar.
    * 
    * @param    eventCode
    *           event code, one of the NOTIFY_* constants
    */
   public void notifyEvent(int eventCode)
   {
      // filter out unwanted events
      if (eventCode != AdminNotify.NOTIFY_LOGOFF &&
          eventCode != AdminNotify.NOTIFY_TARGET)
      {
         return;
      }
      
      // reset the target description
      tgtTarget.setText("");
      tgtError.setText("");
      
      // get the menu
      List<JMenuItem>menuList = menu.getMenuList();
      
      // notify all registered listeners
      if (eventCode == AdminNotify.NOTIFY_LOGOFF)
      {
         tgtServer.setText("");
         tgtSrvSep.setText("");
         
         for(AdminNotify al: logoffList)
         {
            al.notifyEvent(eventCode);
         }

         // rebuild the menu 
         if (menuExt != null)
         {
            // remove the previously inserted custom items, if any
            menuExt.removeCustomMenuItems(menuList);
         }
      }
      else
      {
         for(AdminNotify al: targetList)
         {
            al.notifyEvent(eventCode);
         }
         
         // rebuild the menu 
         if (menuExt == null)
         {
            return;
         }
         
         // remove the previously inserted custom items, if any
         menuExt.removeCustomMenuItems(menuList);
         
         // reinsert the custom items, if targeted
         if (!target.isNone())
         {
            menuExt.insertCustomMenuItems(menuList);
         }
      }
      
      // rebuild the menu
      if (eventCode == AdminNotify.NOTIFY_LOGOFF)
      {
         menu.buildMenu(null);
      }
      else
      {
         menu.buildMenu(adminDef.profile);
      }
      setJMenuBar(menu.getMenu());
      
      if (eventCode == AdminNotify.NOTIFY_LOGOFF)
      {
         // clear aged data
         adminDef = null;
         accountExt = null;
         menuExt = null;
         customAclEditors = null;
         resourceTypes = null;
      }
   }
   
   /** 
    * Creates an errors screen.
    */
   private void errorsScreen()
   {
      JPanel panel = new JPanel();
      panel.setLayout(new BorderLayout());
      
      errData = new DefaultListModel();
      errBox  = new JList(errData);
      
      errScrollPane = new JScrollPane();
      errScrollPane.getViewport().add(errBox);
      
      panel.add(errScrollPane, BorderLayout.CENTER);
      
      JButton back = new JButton("Back");
      JPanel backp = AdminUtils.groupComponents(new Component[][] {{back}});
      panel.add(backp, BorderLayout.SOUTH);
      
      back.addActionListener(new ActionListener() 
         {
            @SuppressWarnings("unused")
            public void actionPerformed(ActionEvent evt) 
            {
               screens.show(swapPanel, showing);
            }
         });
      
      // add this screen to the swapper
      add(panel, SCREEN_ERRORS, this);
   }
   
   /**
    * Finds a parent component that is a Frame. Required for creating dialogs
    * in applets.
    * 
    * @return   instance of parent Frame
    */
   private Frame findParentFrame()
   {
      Component c = getParent();

      while(true)
      {
         if(c instanceof Frame)
         {
            return (Frame)c;
         }

         c = c.getParent();
      }
   }
   
   /**
    * Provides a table data model to represent printable server messages.
    */
   private class PrintMessagesModel
   extends AbstractTableModel
   {
      /** Column names. */
      private final String[] columnNames =
      {
         "Message",
      };
      
      /** Default column sizes. */
      private final int columnSizes[] =
      {
         500
      };
      
      /** Column data types. */
      private final Class<?>[] columnTypes =
      {
         String.class,
      };
      
      /** Core data for the table. */
      private String[] data = null;
   
      /**
       * Construct an instance.
       * 
       * @param    data
       *           The data to provide to the table.
       */
      public PrintMessagesModel(String[] data)
      {
         this.data = data;
      }
      
      /**
       * Gets the number of columns in the table model.
       * 
       * @return   The number of columns.
       */
      public int getColumnCount()
      {
         return columnNames.length;
      }
   
      /**
       * Gets the width of the given column.
       * 
       * @param    col
       *           The column index (0-based).
       * 
       * @return   The width in pixels.
       */
      @SuppressWarnings("unused")
      public int getColumnWidth(int col)
      {
         return columnSizes[col];
      }
   
      /**
       * Gets the number of rows in the table model.
       * 
       * @return   The number of rows.
       */
      public int getRowCount()
      {
         return data.length;
      }
   
      /**
       * Gets the name of a specific column.
       * 
       * @param    col
       *           The column index (0-based).
       * 
       * @return   The column name.
       */
      public String getColumnName(int col)
      {
         return columnNames[col];
      }
   
      /**
       * Get the cell value to be displayed.
       * 
       * @param    col
       *           The column index (0-based).
       * @param    row
       *           The row index (0-based).
       * 
       * @return   The value.
       */
      public Object getValueAt(int row, int col)
      {
         Object datum = data[row];
         return datum;
      }
   
      /**
       * Gets the data type of a specific column.
       * 
       * @param    col
       *           The column index (0-based).
       * 
       * @return   The column type.
       */
      public Class<?> getColumnClass(int col)
      {
         return columnTypes[col];
      }
   }
}
